package org.apache.bcel.verifier.statics;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import org.apache.bcel.Constants;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.DescendingVisitor;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.Node;
import org.apache.bcel.generic.Type;
import org.apache.bcel.verifier.PassVerifier;
import org.apache.bcel.verifier.VerificationResult;
import org.apache.bcel.verifier.Verifier;
import org.apache.bcel.verifier.VerifierFactory;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import org.apache.bcel.verifier.exc.ClassConstraintException;
import org.apache.bcel.verifier.statics.LocalVariablesInfo;
import org.apache.bcel.verifier.statics.StringRepresentation;

public final class Pass2Verifier extends PassVerifier implements Constants {
  private LocalVariablesInfo[] localVariablesInfos;
  
  private final Verifier myOwner;
  
  private static String tostring(Node n) {
    return (new StringRepresentation(n)).toString();
  }
  
  private static boolean validClassMethodName(String name) {
    return validMethodName(name, false);
  }
  
  private static boolean validClassName(String name) {
    Objects.requireNonNull(name, "name");
    return true;
  }
  
  private static boolean validFieldName(String name) {
    return validJavaIdentifier(name);
  }
  
  private static boolean validInterfaceMethodName(String name) {
    if (name.startsWith("<"))
      return false; 
    return validJavaLangMethodName(name);
  }
  
  private static boolean validJavaIdentifier(String name) {
    if (name.isEmpty() || !Character.isJavaIdentifierStart(name.charAt(0)))
      return false; 
    for (int i = 1; i < name.length(); i++) {
      if (!Character.isJavaIdentifierPart(name.charAt(i)))
        return false; 
    } 
    return true;
  }
  
  private static boolean validJavaLangMethodName(String name) {
    return validJavaIdentifier(name);
  }
  
  private static boolean validMethodName(String name, boolean allowStaticInit) {
    if (validJavaLangMethodName(name))
      return true; 
    if (allowStaticInit)
      return (name.equals("<init>") || name.equals("<clinit>")); 
    return name.equals("<init>");
  }
  
  public Pass2Verifier(Verifier owner) {
    this.myOwner = owner;
  }
  
  private void constantPoolEntriesSatisfyStaticConstraints() {
    try {
      JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
      new CPESSC_Visitor(this, jc, null);
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  public VerificationResult do_verify() {
    try {
      VerificationResult vr1 = this.myOwner.doPass1();
      if (vr1.equals(VerificationResult.VR_OK)) {
        this.localVariablesInfos = new LocalVariablesInfo[(Repository.lookupClass(this.myOwner.getClassName()).getMethods()).length];
        VerificationResult vr = VerificationResult.VR_OK;
        try {
          constantPoolEntriesSatisfyStaticConstraints();
          fieldAndMethodRefsAreValid();
          everyClassHasAnAccessibleSuperclass();
          finalMethodsAreNotOverridden();
        } catch (ClassConstraintException cce) {
          vr = new VerificationResult(2, cce.getMessage());
        } 
        return vr;
      } 
      return VerificationResult.VR_NOTYET;
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  private void everyClassHasAnAccessibleSuperclass() {
    try {
      Set<String> hs = new HashSet<>();
      JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
      int supidx = -1;
      while (supidx != 0) {
        supidx = jc.getSuperclassNameIndex();
        if (supidx == 0) {
          if (jc != Repository.lookupClass(Type.OBJECT.getClassName()))
            throw new ClassConstraintException("Superclass of '" + jc
                .getClassName() + "' missing but not " + Type.OBJECT.getClassName() + " itself!"); 
          continue;
        } 
        String supername = jc.getSuperclassName();
        if (!hs.add(supername))
          throw new ClassConstraintException("Circular superclass hierarchy detected."); 
        Verifier v = VerifierFactory.getVerifier(supername);
        VerificationResult vr = v.doPass1();
        if (vr != VerificationResult.VR_OK)
          throw new ClassConstraintException("Could not load in ancestor class '" + supername + "'."); 
        jc = Repository.lookupClass(supername);
        if (jc.isFinal())
          throw new ClassConstraintException("Ancestor class '" + supername + "' has the FINAL access modifier and must therefore not be subclassed."); 
      } 
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  private void fieldAndMethodRefsAreValid() {
    try {
      JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
      DescendingVisitor v = new DescendingVisitor(jc, new FAMRAV_Visitor(this, jc, null));
      v.visit();
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  private void finalMethodsAreNotOverridden() {
    try {
      Map<String, String> map = new HashMap<>();
      JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
      int supidx = -1;
      while (supidx != 0) {
        supidx = jc.getSuperclassNameIndex();
        Method[] methods = jc.getMethods();
        for (Method method : methods) {
          String nameAndSig = method.getName() + method.getSignature();
          if (map.containsKey(nameAndSig)) {
            if (method.isFinal()) {
              if (!method.isPrivate())
                throw new ClassConstraintException("Method '" + nameAndSig + "' in class '" + (String)map.get(nameAndSig) + "' overrides the final (not-overridable) definition in class '" + jc
                    .getClassName() + "'."); 
              addMessage("Method '" + nameAndSig + "' in class '" + (String)map.get(nameAndSig) + "' overrides the final (not-overridable) definition in class '" + jc
                  .getClassName() + "'. This is okay, as the original definition was private; however this constraint leverage was introduced by JLS 8.4.6 (not vmspec2) and the behavior of the Sun verifiers.");
            } else if (!method.isStatic()) {
              map.put(nameAndSig, jc.getClassName());
            } 
          } else if (!method.isStatic()) {
            map.put(nameAndSig, jc.getClassName());
          } 
        } 
        jc = Repository.lookupClass(jc.getSuperclassName());
      } 
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  public LocalVariablesInfo getLocalVariablesInfo(int methodNr) {
    if (verify() != VerificationResult.VR_OK)
      return null; 
    if (methodNr < 0 || methodNr >= this.localVariablesInfos.length)
      throw new AssertionViolatedException("Method number out of range."); 
    return this.localVariablesInfos[methodNr];
  }
}
